//Latest modification on March 25th, 2015 (version 1.23 
//Yuzhen Ye <yye@@indiana.edu> 
//Affiliation: School of Informatics and Computing, Indiana University, Bloomington
#include "crispraln.h"
#include <time.h>

using namespace smallapp;

void printusage(char *errorinfo);

main(int argc, char **argv)
{
	char	inpfile[1000], crtoutfile[1000], noveloutfile[1000], supportedoutfile[1000], crisprfile[1000], repeatfile[1000], listfile[1000];
	int	allowmis = 2;
	double	iden = 0.9;
	int	partial = 0;
	bool	allowredun = false;
	inpfile[0] = noveloutfile[0] = supportedoutfile[0] = crisprfile[0] = listfile[0] = repeatfile[0] = crtoutfile[0] = 0;
	int	num;
	double	numf;

	crispraln	*crispreng = new crispraln;

	for(int i = 0; i < argc; i ++) {
		if((!strcmp(argv[i], "-s")) && argc > i + 1) {
			strcpy(inpfile, argv[i + 1]);
		}
		else if((!strcmp(argv[i], "-l")) && argc > i + 1) {
			strcpy(listfile, argv[i + 1]);
		}
		else if(!strcmp(argv[i], "-partial")) {
			partial = 1;
		}
		else if((!strcmp(argv[i], "-novel")) && argc > i + 1) {
			strcpy(noveloutfile, argv[i + 1]);
		}
		else if((!strcmp(argv[i], "-supported")) && argc > i + 1) {
			strcpy(supportedoutfile, argv[i + 1]);
		}
		else if(!strcmp(argv[i], "-redundant")) {
			allowredun = true;
		}
		else if((!strcmp(argv[i], "-crt")) && argc > i + 1) {
			strcpy(crtoutfile, argv[i + 1]);
		}
		else if((!strcmp(argv[i], "-c")) && argc > i + 1) {
			strcpy(crisprfile, argv[i + 1]);
		}
		else if((!strcmp(argv[i], "-r")) && argc > i + 1) {
			strcpy(repeatfile, argv[i + 1]);
		}
		else if((!strcmp(argv[i], "-d")) && argc > i + 1) {
			sscanf(argv[i + 1], "%d", &allowmis);
			printf("allowmis changed to %d\n", allowmis);
		}
		else if((!strcmp(argv[i], "-p")) && argc > i + 1) {
			allowmis = 100;
			sscanf(argv[i + 1], "%lf", &iden);
			printf("use identity cutoff (instead of allowmis): %lf\n", iden);
		}
		else if(!strcmp(argv[i], "-v")) {
			crispreng -> setverbose();
		}
	}
	if((inpfile[0] == 0) && (listfile[0] == 0)) printusage("Error: input sequence or list file of sequences is not given\n");
	if((crisprfile[0] == 0) && (repeatfile[0] == 0)) printusage("Error: crispr/repeat file is not given\n");
	printf("mark 1\n");	
	time_t jobstart = time(NULL);
	if(crisprfile[0] != 0) {
		crispreng -> readcrispr(crisprfile);
	}
	else {
		crispreng -> readrepeat(repeatfile);
		if(crtoutfile[0] == 0) 
			strcpy(crtoutfile, "tmp.crt");
	}
	
	int	crisprall = 0;
	int	filenum = 0;
	int	findnum;
	if(inpfile[0] != 0) {
		crispreng -> getseqs(inpfile);
		if(crisprfile[0] != 0) {
			findnum = crispreng -> crispralign(allowmis, 1-iden, partial);
			if(noveloutfile[0] != 0) crispreng -> printnovel(noveloutfile);
			if(supportedoutfile[0] != 0) crispreng -> printsupported(supportedoutfile);
		}
		else {
			FILE	*out = smallapp::ckopen(crtoutfile, "w");
			findnum = crispreng -> crisprann(allowmis, 1-iden, partial, allowredun, out);
			fclose(out);
		}
		crisprall += findnum;
		filenum ++;
		printf("%s, find-cripsr %d\n", inpfile, findnum);
	}	
	if(listfile[0] != 0) {
	        FILE    *fp = smallapp::ckopen(listfile, "r");
       	 	char    str[10001], filename[10000];
		FILE	*out = NULL;
		if(crisprfile[0] == 0) {
			out = smallapp::ckopen(crtoutfile, "w");
		}
        	while(fgets(str, 10000, fp))    {
			if(strlen(str) < 1) continue; 
			if(str[0] == '#') continue;
			sscanf(str, "%s", filename);
			printf("filename: %s\n", filename);
			crispreng -> getseqs(filename);
			if(crisprfile[0] != 0) {
				findnum = crispreng -> crispralign(allowmis, 1-iden, partial);
			}
			else {
				findnum = crispreng -> crisprann(allowmis, 1-iden, partial, allowredun, out);
			}
			crisprall += findnum;
			filenum ++;
			printf(">%s cripsr-tot %d\n", filename, findnum);
		}	
		if(crisprfile[0] == 0) {
			fclose(out);
		}
		fclose(fp);
	}

	printf("SUMMARY: Total file %d, total crispr %d\n", filenum, crisprall);
	printf("Results saved to %s\n", crtoutfile);

	delete crispreng;

	time_t jobfinished = time(NULL);
	double  timeused = difftime(jobfinished, jobstart);
	if(timeused < 60) {
		printf("Time used: %d sec\n", int(timeused));
	}
	else {
		printf("Time used: %d min (%.1f hours)\n", int(timeused / 60), timeused / 3600.0);
	}
}

void printusage(char *error)
{
	printf("%s\n", error);
	printf("CrisprAlign version 1.4 -- released on Sep 1, 2015\n");
	printf("CrisprAlign -r RepeatFile <-c CrisprFile> -s SourceSeqFile(fasta) <-l list-file>\n");
	printf("   choose -r or -c, but not both\n");
	printf("   -r Repeat-File: Fasta file of repeats\n");
	printf("   -c Crispr-Annotation-File: output from summarize-crispr.py\n");
	printf("   -s SourceSeqFile (FASTA format): The input sequence  file name\n");
	printf("   -l ListSeqFile (String): a list of input sequences\n");
	printf("  Other optinal parameters:\n");
	printf("   -d integer: mismatch-allowed, default 2\n");
	printf("   -p float: identity cutoff, default 0.9; if set, -d is disabled\n");
	printf("   -partial: incomplete repeats at the ends are allowed, default off\n");
	printf("   -redundant: allow redundant annotation of the same region by multiple reference repeats, default off\n");
	printf("   -v: verbose-on, default off\n");
	printf("  Outout file:\n");
	printf("   -supported file-name: output supported crisprs in a file\n");
	printf("   -novel file-name: output 'novel' crisprts in a file\n");
	printf("   -crt file-name: output identified crisprs from the input sequences to a file\n");
	printf("  Note: use -supported and -novel when crispr file is given in the input, use -crt when repeat file is given\n");
	printf("  Example usage: \n");
	printf("  ./CrisprAlign -r Strep-crt.rep -s SRS013533.contig -partial -d 5 -crt SRS013533.contig.crt\n");

	exit(0);
}
