//Program Name: crispraln

#include <limits.h>
#include <algorithm>
#include "crispraln.h"

using namespace smallapp;
using namespace hashapp;

void crispraln::cleanupseq(void)
{
	int	i;
	for(i = 0; i < num_src; i ++)	{
		delete[] src_seq[i];
		delete[] src_name[i];
	}
	delete[] src_seq;
	delete[] src_name;
	delete[] src_len;
	num_src = 0;
}

void crispraln::cleanupcrispr(void)
{
	int	i;
	for(i = 0; i < num_crispr; i ++)	{
		delete[] crispr_name[i];
		delete[] crispr_con[i];
	}
	delete[] crispr_name;
	delete[] crispr_con;
	if(crispr_seq != NULL) {
		for(i = 0; i < num_crispr; i ++)	
			delete[] crispr_seq[i];
		delete[] crispr_seq;
	}
	if(spacer_idx != NULL) {
		for(i = 0; i < num_crispr; i ++)	{
			delete[] spacer_idx[i];
			delete[] spacer_len[i];
		}
		delete[] spacer_idx;
		delete[] spacer_len;
	}
	if(spacer_len_sum != NULL) 
		delete[] spacer_len_sum;
	if(crispr_len != NULL) 
		delete[] crispr_len;
	if(crispr_truesupport != NULL)
		delete[] crispr_truesupport;
	if(crispr_falsesupport != NULL)
		delete[] crispr_falsesupport;
}

void crispraln::ini_crispraln(void)
{
	src_name = NULL;
	src_seq = NULL;
	src_len = NULL;
	crispr_name = NULL;
	crispr_con = NULL;
	crispr_seq = NULL;
	crispr_len = NULL;
	crispr_seq = NULL;
	repeat_seq = NULL;
	repeat_len = NULL;
	crispr_truesupport = NULL;
	crispr_falsesupport = NULL;	
	spacer_len_sum = NULL;
	spacer_idx = NULL;
	spacer_len = NULL;
	spacerDiff = 0.5;
	minSpacerLen = 10;
	maxSpacerLen = 60;
	minRepeatCopy = 2;
	verbose = 0;
}

//get reads from file
void crispraln::getseqs(char *seqfile)
{
	if(src_name != NULL) cleanupseq();
	//get read num first
	FILE	*fp0 = smallapp::ckopen(seqfile, "r");
	int	n = getnumread(fp0);
	fclose(fp0);

	src_name = new char*[2 * n];
	src_seq = new char*[2 * n];
	src_len = new int[2 * n];
	int	i, j;
	for(i = 0; i < 2 * n; i ++) src_len[i] = 0;

	//now get reads
	FILE* 	fp = smallapp::ckopen(seqfile, "r");
	num_src = readseq(src_seq, src_len, src_name, fp);
	fclose(fp);

	//Considering the reverse complmentary sequences
	/*
	for(i = 0; i < num_src; i ++)	{
		src_len[num_src + i] = src_len[i];
		src_seq[num_src + i] = new char[src_len[i]];
		src_name[num_src + i] = new char[strlen(src_name[i]) + 15];
		sprintf(src_name[num_src + i], "%s reverse comp", src_name[i]);
		for(j = 0; j < src_len[i]; j ++)	{
			src_seq[num_src + i][j] = rev(src_seq[i][src_len[i] - 1 - j]);
		}
	}
	num_src = 2 * num_src;
	*/

	int	max_seqlen = src_len[0];
	int	min_seqlen = 10000;
	for(int i = 1; i < num_src; i ++) {
		if(src_len[i] > max_seqlen) max_seqlen = src_len[i];
		if(min_seqlen > src_len[i]) min_seqlen = src_len[i];
	}
	printf("totseq %d, max_seqlen %d, min_seqlen %d\n", num_src, max_seqlen, min_seqlen);
	fflush(stdout);
}

int crispraln::readseq(char **seq, int *len_seq, char **src_name, FILE *fp)
{
	int	i, j, num_src0;
	char	str[10001], tmp[1000];
	char	*src_seq = new char[MAX_SEQLENGTH];

	num_src0 = 0;
	while(fgets(str, 10000, fp))	{
		if(str[0] == '>')	{
			sscanf(&str[1], "%s", tmp); //v1.
			src_name[num_src0] = new char[strlen(tmp) + 1]; //v1.1
			strcpy(src_name[num_src0], tmp); //v1.1
			if(num_src0 == 0)	{
				num_src0 ++;
				continue;
			}
			if(len_seq[num_src0 - 1] < kmer)	{
				len_seq[num_src0 - 1] = 0;
				continue;
			}
			seq[num_src0 - 1] = new char[len_seq[num_src0 - 1]];
			for(i = 0; i < len_seq[num_src0 - 1]; i ++)	{
				seq[num_src0 - 1][i] = src_seq[i];
			}
			num_src0 ++;
		} else	{
			for(i = 0; i < strlen(str); i ++)	{
				if(str[i] >= 'A' && str[i] <= 'Z' || str[i] >= 'a' && str[i] <= 'z')	{
					src_seq[len_seq[num_src0 - 1] ++] = smallapp::char2int(str[i]);
				}
			}
		}
	}
	if(len_seq[num_src0 - 1] >= kmer)	{
		seq[num_src0 - 1] = new char[len_seq[num_src0 - 1]];
		for(i = 0; i < len_seq[num_src0 - 1]; i ++)	{
			seq[num_src0 - 1][i] = src_seq[i];
		}
	} else	{
		num_src0 --;
	}
	delete[] src_seq;
	return(num_src0);
}

void crispraln::readrepeat(char *repeatfile)
{
	FILE	*fp0 = smallapp::ckopen(repeatfile, "r");
	int	n = smallapp::getnumread(fp0);
	printf("expected repeats: %d\n", n);
	fclose(fp0);
	num_crispr = 0;
	crispr_name = new char*[n];
	crispr_con = new char*[n]; //in char
	repeat_seq = new char*[n]; //in int
	repeat_len = new int[n];
	int	i, j;
	for(i = 0; i < n; i ++) {
		repeat_len[i] = 0;
	}
	FILE	*fp = smallapp::ckopen(repeatfile, "r");
	int	num_crispr0 = 0;
	char	str[100001], tmp[1000];
	str[0] = 0;
	while(fgets(str, 10000, fp))	{
		if(str[0] == '>')	{
			sscanf(&str[1], "%s", tmp);
			crispr_name[num_crispr0] = new char[strlen(tmp) + 1];
			strcpy(crispr_name[num_crispr0], tmp);
			num_crispr0 ++;
		}
		else if((strlen(str) > 0) && (repeat_len[num_crispr0 - 1] == 0)) { //YY, Sep 1, 2015
			crispr_con[num_crispr0 - 1] = new char[strlen(str) + 1];
			sscanf(str, "%s", crispr_con[num_crispr0 - 1]); //YY, Sep 1, 2015
			repeat_seq[num_crispr0 - 1] = new char[strlen(str) + 1];
			for(i = 0; i < strlen(str); i ++) { //current repeat-consensus
                                if(str[i] >= 'A' && str[i] <= 'Z' || str[i] >= 'a' && str[i] <= 'z')    {
                                        repeat_seq[num_crispr0 - 1][repeat_len[num_crispr0 - 1] ++] = smallapp::char2int(str[i]);
				}
			}
		} 
		str[0] = 0;
	}
	num_crispr = num_crispr0;
	printf("total repeats read-in %d\n", num_crispr0);
	//char *seq = int2seq(repeat_seq[0], repeat_len[0]);
	//printf("seq: %s\n", seq);
	fclose(fp);
}

void crispraln::readcrispr(char *annfile)
{
	//get read num first
	FILE	*fp0 = smallapp::ckopen(annfile, "r");
	int	n = smallapp::getnumread(fp0);
	fclose(fp0);
	printf("expect crispr number %d\n", n);
	num_crispr = 0;
	crispr_name = new char*[n];
	crispr_con = new char*[n]; //in char
	crispr_seq = new char*[n];
	crispr_len = new int[n];
	spacer_idx = new int*[n];
	spacer_len = new int*[n];
	spacer_num = new int[n];
	repeat_seq = new char*[n]; //in int
	repeat_len = new int[n];
	int	i, j;
	for(i = 0; i < n; i ++) {
		repeat_len[i] = crispr_len[i] = spacer_num[i] = 0;
	}

	char	*seq_tmp = new char[MAX_SEQLENGTH];
	seq_tmp[0] = 0;

	FILE	*fp = smallapp::ckopen(annfile, "r");
	int	num_crispr0 = 0;
	char	str[100001], tmp[1000], tmp2[1000];
	int	spacernum0 = 0;
	int	spaceridx[10000], spacerlen[10000];
	while(fgets(str, 10000, fp))	{
		if(str[0] == '>')	{
			sscanf(&str[1], "%s%s", tmp, tmp2);
			crispr_name[num_crispr0] = new char[strlen(tmp) + 1];
			crispr_con[num_crispr0] = new char[strlen(tmp2) + 1];
			strcpy(crispr_name[num_crispr0], tmp);
			strcpy(crispr_con[num_crispr0], tmp2);
			repeat_seq[num_crispr0] = new char[strlen(tmp2) + 1];
			for(i = 0; i < strlen(tmp2); i ++) { //current repeat-consensus
                                if(tmp2[i] >= 'A' && tmp2[i] <= 'Z' || tmp2[i] >= 'a' && tmp2[i] <= 'z')    {
                                        repeat_seq[num_crispr0][repeat_len[num_crispr0] ++] = smallapp::char2int(tmp2[i]);
				}
			}
			if(num_crispr0 == 0)	{
				num_crispr0 ++;
				continue;
			}
			//assign previous crispr
			crispr_seq[num_crispr0 - 1] = new char[crispr_len[num_crispr0 - 1]];
			spacernum0 = 0;
			for(i = 0; i < crispr_len[num_crispr0 - 1]; i ++)	{ //note, num_crispr0 - 1
				crispr_seq[num_crispr0 - 1][i] = smallapp::char2int(seq_tmp[i]);
				if((seq_tmp[i] >= 'a' && seq_tmp[i] <= 'z') && (seq_tmp[i - 1] >= 'A' && seq_tmp[i - 1] <= 'Z')) {
					spacerlen[spacernum0] = 1;
					spaceridx[spacernum0 ++] = i;
					//printf("spacerid %d, %c %c, start-at %d, len %d\n", spacernum0 - 1, seq_tmp[i], seq_tmp[i - 1], spaceridx[spacernum0 - 1], spacerlen[spacernum0 - 1]);
				}
				else if(seq_tmp[i] >= 'a' && seq_tmp[i] <= 'z') {
					spacerlen[spacernum0 - 1] ++;
				}
			}
			spacer_num[num_crispr0 - 1] = spacernum0;
			spacer_idx[num_crispr0 - 1] = new int[spacernum0];
			spacer_len[num_crispr0 - 1] = new int[spacernum0];
			for(i = 0; i < spacernum0; i ++) {
				spacer_idx[num_crispr0 - 1][i] = spaceridx[i];	
				spacer_len[num_crispr0 - 1][i] = spacerlen[i];
			}
			num_crispr0 ++;
			seq_tmp[0] = 0;
		} else	{
			for(i = 0; i < strlen(str); i ++)	{
				if(str[i] >= 'A' && str[i] <= 'Z' || str[i] >= 'a' && str[i] <= 'z')	{
					//printf("crispr_len of %d, %d\n", num_crispr0 - 1, crispr_len[num_crispr0 - 1]);
					seq_tmp[crispr_len[num_crispr0 - 1] ++] = str[i];
				}
			}
		}
	}
	fclose(fp);
	crispr_seq[num_crispr0 - 1] = new char[crispr_len[num_crispr0 - 1]];
	spacernum0 = 0;
	for(i = 0; i < crispr_len[num_crispr0 - 1]; i ++)	{ //note, num_crispr0 - 1
		crispr_seq[num_crispr0 - 1][i] = smallapp::char2int(seq_tmp[i]);
		if((seq_tmp[i] >= 'a' && seq_tmp[i] <= 'z') && (seq_tmp[i - 1] >= 'A' && seq_tmp[i - 1] <= 'Z')) {
			spacerlen[spacernum0] = 1;
			spaceridx[spacernum0 ++] = i;
			//printf("spacerid %d, %c %c, start-at %d, len %d\n", spacernum0 - 1, seq_tmp[i], seq_tmp[i - 1], spaceridx[spacernum0 - 1], spacerlen[spacernum0 - 1]);
		}
		else if(seq_tmp[i] >= 'a' && seq_tmp[i] <= 'z') {
			spacerlen[spacernum0 - 1] ++;
		}
	}
	spacer_num[num_crispr0 - 1] = spacernum0;
	spacer_idx[num_crispr0 - 1] = new int[spacernum0];
	spacer_len[num_crispr0 - 1] = new int[spacernum0];
	for(i = 0; i < spacernum0; i ++) {
		spacer_idx[num_crispr0 - 1][i] = spaceridx[i];	
		spacer_len[num_crispr0 - 1][i] = spacerlen[i];
	}

	delete[] seq_tmp;

	num_crispr = num_crispr0;
	spacer_len_sum = new int[num_crispr];
	for(i = 0; i < num_crispr; i ++) {
		spacer_len_sum[i] = 0;
		for(j = 0; j < spacer_num[i]; j ++) {
			spacer_len_sum[i] += spacer_len[i][j];
		}
	}
	printf("total crispr: %d\n", num_crispr);
}

//-----------------------------------------------------------------------------------------------
int crispraln::spacersim(int r, int s, int *repeatpos, int repeatcopy)
{
	int	m, k, p, l;
	int	maxdiagsim, maxdiagidx, maxdiaglen, diagsim, diag, diaglen; 
	int	mis, alnlen, same;
	maxdiagsim = 0;
	int	band = 10;
	vector<int> aln;
	for(diag = 0; diag < spacer_num[r]; diag ++) { //check by diagnals
		diagsim = diaglen = 0;
		for(m = 0; m < repeatcopy - 1; m ++) { //repeat copy m in repeat-cluster c, found in sequence s
			k = m + diag;
			if(k >= spacer_num[r]) continue;
			//crispr r, spacer k
			p = repeatpos[m] + repeat_len[r];
			l = repeatpos[m + 1] - p; 
			mis = global_align(crispr_seq[r] + spacer_idx[r][k], src_seq[s] + p, spacer_len[r][k], l, band, aln, alnlen, same);
			diagsim += same;
			diaglen += alnlen;
		}
		if(diag == 0 || diagsim > maxdiagsim) {
			maxdiagsim = diagsim;
			maxdiagidx = diag;
			maxdiaglen = diaglen;
		}
	}
	if(verbose) {
		printf("maxdiagidx %d\n", maxdiagidx);
		for(m = 0; m < repeatcopy - 1; m ++) { //repeat copy m in repeat-cluster c, found in sequence s
			k = m + maxdiagidx;
			if(k >= spacer_num[r]) continue;
			//crispr r, spacer k
			p = repeatpos[m] + repeat_len[r];
			l = repeatpos[m + 1] - p; 
			mis = global_align(crispr_seq[r] + spacer_idx[r][k], src_seq[s] + p, spacer_len[r][k], l, band, aln, alnlen, same);
			printf("**align spacers: crispr %d spacer-start %d (len %d) with seq %d (len %d), result same %d, mis %d, alnlen %d\n", 
				r, spacer_idx[r][k], spacer_len[r][k], p, l, same, mis, alnlen);
			//writeseq(crispr_seq[r] + spacer_idx[r][k], "spacer-in-input-crispr", spacer_len[r][k], stdout);
			//writeseq(src_seq[s] + p, "spacer-identified", l, stdout);
			printaln(crispr_seq[r] + spacer_idx[r][k], spacer_len[r][k], src_seq[s] + p, l, aln);
			diagsim += same;
		}
	}
	double	sim = maxdiagsim * 1.0 / maxdiaglen;
	printf("  Total spacer %d maxdiag %d maxdiagsim %d alnlen %d per %.2f\n", repeatcopy - 1, maxdiagidx, maxdiagsim, maxdiaglen, sim);

	return sim;
}

int crispraln::left_partial_repeat(int repeatlen, char *repeatseq, int firstcopypos, int srclen, char *srcseq)
{
	int	p = firstcopypos;
	int	maxpartial_mis = 1;
	//int	minpartial = repeatlen / 2;
	int	minpartial = 10;
	if((p > minSpacerLen + minpartial) && (p < maxSpacerLen + repeatlen)) {
		for(int lenshort = 1; lenshort <= repeatlen - minpartial; lenshort ++) {
			int	miss = 0;
			for(int k = 0; k < repeatlen - lenshort; k ++) {
				if(srcseq[k] != repeatseq[k + lenshort]) miss ++;
				if(miss > maxpartial_mis) break; //
			}
			if(miss <= maxpartial_mis) {
				return lenshort;
			}
		}
	}
	return 0;
}

int crispraln::right_partial_repeat(int repeatlen, char *repeatseq, int lastcopypos, int srclen, char *srcseq)
{
	int	p = srclen - (lastcopypos + repeatlen);
	//int	minpartial = repeatlen / 2;
	int	minpartial = 10;
	int	maxpartial_mis = 1;
	if((p > minSpacerLen + minpartial) && (p < maxSpacerLen + repeatlen)) {
		for(int lenshort = 1; lenshort <= repeatlen - minpartial; lenshort ++) {
			int miss = 0;
			int	anchor = srclen - repeatlen - 1 + lenshort;
			for(int k = 0; k < repeatlen - lenshort; k ++) {
				if(srcseq[anchor + k] != repeatseq[k]) miss ++;
				if(miss > maxpartial_mis) break; //
			}
			if(miss <= maxpartial_mis) {
				return lenshort;
				break;
			}
		}
	}
	return 0;
}

void crispraln::consensus_new(int len)
{
	for(int i = 0; i < len; i ++) {
		cA[i] = cT[i] = cC[i] = cG[i] = 0;
	} 
}

void crispraln::consensus_add(int len, char *seq)
{
	for(int i = 0; i < len; i ++) {
		if(seq[i] == '-') continue;
		else if(seq[i] == 'A') cA[i] ++;
		else if(seq[i] == 'T') cT[i] ++;
		else if(seq[i] == 'C') cC[i] ++;
		else if(seq[i] == 'G') cG[i] ++;
		else {
			printf("unknown base #%c#\n", seq[i]);
			exit(0);
		}
	}
}
void crispraln::consensus_get(int len, char *con)
{
	char	cbase;
	int	i;
	con[0] = 0; //YY, Feb 6, 2015
	for(i = 0; i < len; i ++) {
		if((cA[i] >= cC[i]) && (cA[i] >= cT[i]) && (cA[i] >= cG[i])) cbase = 'A';
                else if((cC[i] >= cA[i]) && (cC[i] >= cT[i]) && (cC[i] >= cG[i])) cbase = 'C';
		else if((cT[i] >= cC[i]) && (cT[i] >= cA[i]) && (cT[i] >= cG[i])) cbase = 'T';
                else cbase = 'G';
		con[i] = cbase;
	}
	con[i] = '\0';
}

int crispraln::get_mismatch(int repeatlen, char *repeat_int, char *repeat_char)
{
	int	mis = 0;	
	for(int i = 0; i < repeatlen; i ++) {
		int j = smallapp::char2int(repeat_char[i]);
		if(repeat_int[i] != j) mis ++;
	}
	return mis;
}

//-----------------------------------------------------------------------------------------------
int crispraln::crisprann(int allowmis, double iden, int partial, bool allowredun, FILE *out)
{
	int	s, r, c, m, p, t, total, shorter;
	int	pos[10000], repeatidx[10000], repeatcopy[10000];
	int	howmany;
	int	findcrispr = 0;
	char	rseq[1000], sseq[1000], curr_cons[1000], sname[1000];
	int	crisprnum = 0;
	CRISPR	*crispr = new CRISPR[100];
	int	thiscopy;
	int	slen, rep, rlen;
	char	**thisseq = new char*[2];
	int	mis = allowmis;
	for(s = 0; s < num_src; s ++) {
		//printf("\nCheck seq %d %s len %d\n", s, src_name[s], src_len[s]);
		//Feb 24, 2015, YY
		//thiseq[0] -- forward strand; thisseq[1] -- reverse strand
		slen = src_len[s];
		for(t = 0; t < 2; t ++) thisseq[t] = new char[slen + 1];
		for(p = 0; p < slen; p ++)	{
			thisseq[0][p] = src_seq[s][p];
			thisseq[1][p] = rev(src_seq[s][slen - 1 - p]);
		}
		thisseq[0][slen] = thisseq[1][slen] = 0;
		total = 0;
		for(t = 0; t < 2; t ++) {
		   	for(r = 0; r < num_crispr; r ++) {
				rlen = repeat_len[r];
				if(allowmis == 100) mis = floor(iden * rlen + 0.5); 
				howmany = findrepeat(rlen, repeat_seq[r], slen, thisseq[t], mis, 0, pos, repeatidx, repeatcopy); 
				//printf("Compare with cripsr %d %s repeat-len %d, repeatfind %d copy[0] %d\n", r, crispr_name[r], rlen, howmany, repeatcopy[0]);
				if(howmany < 1) continue;
				for(c = 0; c < howmany; c ++) {
					thiscopy = 0;
					consensus_new(rlen);
					for(m = 0; m < repeatcopy[c]; m ++) {
						int2seq(thisseq[t], slen, pos[repeatidx[c] + m], rlen, rseq); 
						consensus_add(rlen, rseq);
					}
					consensus_get(rlen, rseq);
					char *rseq_int = seq2int(rseq, rlen);	
					if(partial && (c == 0)) {
						shorter = left_partial_repeat(rlen, rseq_int, pos[repeatidx[c]], src_len[s], thisseq[t]);
						if(shorter) crispr[total].repeatpos[thiscopy ++] = - shorter;
					}
					for(m = 0; m < repeatcopy[c]; m ++) crispr[total].repeatpos[thiscopy ++] = pos[repeatidx[c] + m];
					if(partial && (c == howmany - 1)) {
						shorter = right_partial_repeat(rlen, rseq_int, pos[repeatidx[c] + repeatcopy[c] - 1], slen, thisseq[t]);
						if(shorter) crispr[total].repeatpos[thiscopy ++] = slen - rlen - 1 + shorter;
					}
					delete[] rseq_int;
					if(thiscopy < minRepeatCopy) continue;

					consensus_get(repeat_len[r], curr_cons);
					crispr[total].mismatch = get_mismatch(rlen, repeat_seq[r], curr_cons);
					crispr[total].beg = crispr[total].repeatpos[0];
					crispr[total].end = crispr[total].repeatpos[thiscopy - 1] + rlen;
					crispr[total].copynum = thiscopy;
					crispr[total].strand = t;
					crispr[total].valid = true;
					crispr[total].repeat = r;
					crispr[total].repeatlen = rlen;
					//printf("find crispr %d, copynum %d, beg %d, end %d\n", total, thiscopy, crispr[total].beg, crispr[total].end);
					total ++;
			    	}
			}
		}

		if(!total) continue;
		int	totalvalid = total;
		if((total > 1) && (!allowredun)) {
			//sort crispr arrays
			std::qsort(crispr, total, sizeof(CRISPR), smallapp::comparcrispr);
			//for(r = 0; r < total; r ++) {
			//	printf("crispr %d repeatlen %d mismatch %d\n", r, crispr[r].repeatlen, crispr[r].mismatch);
			//}
			int	span, totlen;
			for(r = 0; r < total; r ++) {
				//check if this cripr has already been found using a previous repeat
				bool overlap = false;
				for(m = 0; m < r; m ++) {
					if(crispr[m].valid == false) continue;
					if(crispr[r].strand == crispr[m].strand) { //same strand
						span = max(crispr[m].end, crispr[r].end) - min(crispr[m].beg, crispr[r].beg); 
					}
					else { //opposite strands
						int mbeg = slen - crispr[m].beg;
						int mend = slen - crispr[m].end;
						span = max(mend, crispr[r].end) - min(mbeg, crispr[r].beg); 
					}
					totlen = (crispr[m].end - crispr[m].beg + 1) + (crispr[r].end - crispr[r].beg + 1);
					if(span < totlen - 5) { //version 1.2
						overlap = true;
						break;
					}
				}
				if(overlap) { 
					crispr[r].valid = false; 
					totalvalid --;
				}
			}
		}
		//printf("total %d, valid %d\n", total, totalvalid);

		//output crispr arrays
		for(t = 0; t < 2; t ++) {
			int add = 0;
			for(r = 0; r < total; r ++) {
				if((crispr[r].strand == t) && crispr[r].valid)
					add ++;
			}
			if(add == 0) continue;
			if(t == 0) sprintf(sname, "%s", src_name[s]);
			else sprintf(sname, "%s reverse comp", src_name[s]);
			fprintf(out, "SEQ: %s length %d\n", sname, slen);
			fprintf(out, "Bases: %d\n\n\n", slen);
			for(r = 0; r < total; r ++) {
				rep = crispr[r].repeat;
				rlen = repeat_len[rep];
				if((crispr[r].strand != t) || (!crispr[r].valid)) continue;
				fprintf(out, "CRISPR %d  Range: %d - %d\n", r, crispr[r].beg, crispr[r].end);
				fprintf(out, "POSITION        REPEAT                          SPACER\n");
				consensus_new(rlen);
				fprintf(out, "--------------------------------------------------------------\n");
				int	spacer_ave = 0;
				for(int p = 0; p < crispr[r].copynum; p ++) {
					int2seq(thisseq[t], slen, crispr[r].repeatpos[p], rlen, rseq); 
					consensus_add(rlen, rseq);
					if(p == crispr[r].copynum - 1) {
						fprintf(out, "%d\t\t%s\n", crispr[r].repeatpos[p], rseq);
					}
					else {
						int2seq(thisseq[t], slen, crispr[r].repeatpos[p] + rlen, crispr[r].repeatpos[p + 1] - crispr[r].repeatpos[p] - rlen, sseq); 
						fprintf(out, "%d\t\t%s\t%s\n", crispr[r].repeatpos[p], rseq, sseq);
					}
					if(p < thiscopy - 1) spacer_ave += crispr[r].repeatpos[p + 1] - crispr[r].repeatpos[p] - rlen;
				}
				fprintf(out, "--------------------------------------------------------------\n");
				consensus_get(rlen, curr_cons);
				//printf("Repeat-con\t%s\n", curr_cons);
				fprintf(out, "Repeat-con\t%s\n", curr_cons);
				fprintf(out, "Repeats: %d\tAverage Length: %d\tAverage Length: %d\n", crispr[r].copynum, rlen, spacer_ave / (crispr[r].copynum - 1));
				fprintf(out, "==%s instance %d con %s mismatch %d (ref: %s)\n\n\n", sname, totalvalid, curr_cons, get_mismatch(rlen, repeat_seq[rep], curr_cons), crispr_name[rep]);
				findcrispr ++;
			}
		}
		for(c = 0; c < 2; c ++) delete[] thisseq[c];
	}

	return findcrispr;
}

//-----------------------------------------------------------------------------------------------
int crispraln::crispralign(int allowmis, double iden, int partial)
{
	int	r, s, c, k, m, howmany, p, l, thiscopy, shorter;
	int	pos[10000], repeatidx[10000], repeatcopy[10000], repeatpos[10000];;
	int	falsesupport_list[10000], truesupport_list[10000];
	int	findcrispr = 0;
	int	mis = allowmis;
printf("mark 1\n");
	crispr_falsesupport = new int[num_crispr];
	crispr_truesupport = new int[num_crispr];

	for(r = 0; r < num_crispr; r ++) {
		printf("Now check crispr: %d (id %s), repeat-len %d, spacer_num %d\n", r, crispr_name[r], repeat_len[r], spacer_num[r]);
		int	falsesupport = 0;
		int	truesupport = 0;
		for(s = 0; s < num_src; s ++) {
			if(allowmis == 100) mis = repeat_len[r] * iden;
			howmany = findrepeat(repeat_len[r], repeat_seq[r], src_len[s], src_seq[s], allowmis, 1, pos, repeatidx, repeatcopy); 
			if(howmany < 1) continue;
			printf("Compare cripsr %d with seq %d id:%s (len %d)\n", r, s, src_name[s], src_len[s]);
			//printf("  total identified repeat-cluster %d\n", howmany);
			for(c = 0; c < howmany; c ++) {
				thiscopy = 0;
				if(partial && (c == 0)) {
					shorter = left_partial_repeat(repeat_len[r], repeat_seq[r], pos[repeatidx[c]], src_len[s], src_seq[s]);
					if(shorter) repeatpos[thiscopy ++] = - shorter;
				}
				for(m = 0; m < repeatcopy[c]; m ++) repeatpos[thiscopy ++] = pos[repeatidx[c] + m];
				if(partial && (c == howmany - 1)) {
					shorter = right_partial_repeat(repeat_len[r], repeat_seq[r], pos[repeatidx[c] + repeatcopy[c] - 1], src_len[s], src_seq[s]);
					if(shorter) repeatpos[thiscopy ++] = src_len[s] - repeat_len[r] - 1 + shorter;
				}
				if(thiscopy < minRepeatCopy) continue;
				printf("  Check repeat-cluster %d with repeat-copy %d\n", c, repeatcopy[c]);
				double sim = spacersim(r, s, repeatpos, thiscopy);
				if(sim > 0.7) {
					printf("  Bad candidate, repeat copy %d, with similar spacers\n", repeatcopy[c]);
					falsesupport_list[falsesupport ++] = s;
				}
				else {
					printf("  Good candidate, repeat copy %d, with disimilar spacers\n", repeatcopy[c]);
					findcrispr ++;	
					truesupport_list[truesupport ++] = s;
				}
			}
		}
		crispr_falsesupport[r] = falsesupport;
		crispr_truesupport[r] = truesupport;
	}

	int	total_withtruesupport = 0;
	int	total_withfalsesupport = 0;
	for(r = 0; r < num_crispr; r ++) {
		if(crispr_falsesupport[r] > 0) total_withfalsesupport ++;
		else if(crispr_truesupport[r] > 0) total_withtruesupport ++;
	}
	int	total_novel = num_crispr - total_withfalsesupport - total_withtruesupport;
	printf(">>Summary: total crispr %d, with-false-support (similar spacer) %d, with-true-support (dissimlar space) %d, novel %d\n", 
		num_crispr, total_withfalsesupport, total_withtruesupport, total_novel);

	return findcrispr;
}

void crispraln::printsupported(char *file0)
{
	char	filename1[1000], filename[1000];
	sprintf(filename, "%s.list", file0);
	sprintf(filename1, "%s.seq", file0);
	FILE	*fp, *fp1;
	fp = smallapp::ckopen(filename, "w");
	fp1 = smallapp::ckopen(filename1, "w");
	fprintf(fp, ">>With true support\n");
	fprintf(fp, ">>ID   repeat-len repeat-copy num-seq-in-which-the-crispr-is-found-with-dissimilar-spacer\n");
	int	r;
	for(r = 0; r < num_crispr; r ++) {
		if(crispr_falsesupport[r] > 0) continue;
		if(crispr_truesupport[r] <= 0) continue;
		fprintf(fp, "%s %d %d %d\n", crispr_name[r], repeat_len[r], spacer_num[r] + 1, crispr_truesupport[r]); 	
		fprintf(fp1, ">%s %d %d %d\n%s\n", crispr_name[r], repeat_len[r], spacer_num[r] + 1, crispr_truesupport[r], crispr_con[r]);
	}
	fclose(fp);
	fclose(fp1);
}

void crispraln::printnovel(char *file0)
{
	char	filename1[1000], filename[1000];
	sprintf(filename, "%s.list", file0);
	sprintf(filename1, "%s.seq", file0);
	FILE	*fp, *fp1;
	fp = smallapp::ckopen(filename, "w");
	fp1 = smallapp::ckopen(filename1, "w");
	fprintf(fp, ">>Novel crisprs?\n");
	fprintf(fp, ">>ID   repeat-len repeat-copy\n"); 
	int	r;
	for(r = 0; r < num_crispr; r ++) {
		if(crispr_falsesupport[r] > 0) continue;
		if(crispr_truesupport[r] > 0) continue;
		fprintf(fp, "%s %d %d\n", crispr_name[r], repeat_len[r], spacer_num[r] + 1);
		fprintf(fp1, ">%s %d %d\n%s\n", crispr_name[r], repeat_len[r], spacer_num[r] + 1, crispr_con[r]);
	}
	fclose(fp);
	fclose(fp1);


}

//-----------------------------------------------------------------------------------------------
int crispraln::findrepeat(int repeatlen, char *repeatseq, int seqlen, char *seqseq, int allowmis, int removesmall, int* pos, int *repeatidx, int *repeatcopy)
{
	//using sliding window
	int	mis = 0;
	int	howmanyrepeat = 0;
	int	i, j;
	for(i = 0; i < seqlen - repeatlen; i ++) {
		mis = 0;
		for(j = 0; j < repeatlen; j ++) {
			if(seqseq[i + j] != repeatseq[j]) mis ++;
			if(mis > allowmis) break;
		}
		if(mis <= allowmis) {
			pos[howmanyrepeat ++] = i;
			i += repeatlen + minSpacerLen;
			//printf("repeat %d, start-pos %d\n", howmanyrepeat - 1, i);
		}
	}
	if(howmanyrepeat < 1) return 0;

	//define clusters of repeats
	int	howmanyclust = 1;
	repeatidx[0] = 0; 
	repeatcopy[0] = 1;
	for(i = 1; i < howmanyrepeat; i ++) {
		int	spacerlen = pos[i] - pos[i - 1] - repeatlen;
		//printf("repeat %d-%d, spacerlen %d, maxSpacerLen %d\n", i - 1, i, spacerlen, maxSpacerLen); 
		if(pos[i] - pos[i - 1] - repeatlen > maxSpacerLen) {
			repeatidx[howmanyclust] = i;
			repeatcopy[howmanyclust] = 1;
			howmanyclust += 1;
		}
		else {
			repeatcopy[howmanyclust - 1] ++;
		}
	}
	//remove small clusters of repeats
	int	howmanyclust0 = howmanyclust;
	howmanyclust = 0;
	for(i = 0; i < howmanyclust0; i ++) {
		if(removesmall && (repeatcopy[i] < minRepeatCopy)) continue;
		repeatidx[howmanyclust] = repeatidx[i];	
		repeatcopy[howmanyclust] = repeatcopy[i];
		howmanyclust ++;
	}
	return howmanyclust;
}

void crispraln::printaln(char *seq1, int len1, char *seq2, int len2, vector<int> aln)
{
	char	*alnseq1 = new char[len1 + len2 + 1];
	char	*alnseq2 = new char[len1 + len2 + 1];
	char	*alnstr = new char[len1 + len2 + 1];
	alnseq1[0] = alnseq2[0] = alnstr[0]= 0;
	int	p1, p2, i, a;
	p1 = p2 = 0;
	//writeseq(seq1, "seq1", len1, stdout);
	//writeseq(seq2, "seq2", len2, stdout);
	for (i = 0; i < aln.size(); i ++) {
		a = aln[i];
		if(a == 0 || a == 1) {
			sprintf(alnseq1, "%s%c", alnseq1, na_name[seq1[p1 ++]] - 'a' + 'A');
			sprintf(alnseq2, "%s%c", alnseq2, na_name[seq2[p2 ++]] - 'a' + 'A');
			if(a == 0) strcat(alnstr, " ");
			else strcat(alnstr, "x");
		}
		else if(a == 2) {
			sprintf(alnseq1, "%s%c", alnseq1, na_name[seq1[p1 ++]] - 'a' + 'A');
			strcat(alnseq2, "-");
			strcat(alnstr, "x");
		}
		else {
			strcat(alnseq1, "-");
			sprintf(alnseq2, "%s%c", alnseq2, na_name[seq2[p2 ++]] - 'a' + 'A');
			strcat(alnstr, "x");
		}
	}
	printf("%-8s%s\n", "aln1", alnseq1);
	printf("%-8s%s\n", " ", alnstr);
	printf("%-8s%s\n", "aln2", alnseq2);
	delete[] alnseq1;
	delete[] alnseq2;
	delete[] alnstr;
}

void crispraln::int2seq(char *src_seq, int srclen, int start, int len, char *seq)
{
	int	add = 0;
	for(int i = start; i < start + len; i ++)
		if((i < 0) || (i >= srclen)) seq[add ++] = '-'; 
		else seq[add ++] = na_name[src_seq[i]] - 'a' + 'A';
	seq[add] = 0;
}

void crispraln::displayseq(char *src_seq, int len)
{
	for(int i = 0; i < len; i ++)
		printf("%c", na_name[src_seq[i]] - 'a' + 'A');
	printf("\n");
}

char* crispraln::int2seq(char *src_seq, int len)
{
	char    *seq = new char[len + 1];
	for(int i = 0; i < len; i ++)
		seq[i] = na_name[src_seq[i]] - 'a' + 'A';
	seq[len] = 0;
	return seq;
}

char* crispraln::seq2int(char *src_seq, int len)
{
	char    *seq = new char[len + 1];
	for(int i = 0; i < len; i ++)
		seq[i] = smallapp::char2int(src_seq[i]);
	seq[len] = 0;
	return seq;
}

void crispraln::writeseq(char *src_seq, char *src_name, int len, FILE *fp)
{
	int	i, j, k;

	fprintf(fp, ">%s\n", src_name);
	for(i = 0; i < len; i ++)	{
		fprintf(fp, "%c", na_name[src_seq[i]] - 'a' + 'A');
		if(i % 60 == 59)	{
			fprintf(fp, "\n");
		}
	}
	if(i % 60 != 0)	{
		fprintf(fp, "\n");
	}
}
